const EventTarget = require('./eventTarget.js');
const { extract } = require('./objectManipulation');

const SERVER_MESSAGE_MARKER = '@@FROM_SERVER';

const REDUX_INIT_ACTION = '@@INIT';
const SET_STATE_ACTION = '@@SET_STATE';

class SyncClient extends EventTarget {
  constructor({serverUrl = 'ws://localhost:9090', room = 'global', synchronizedActions = [], synchronizedState = null } = {}) {
    super();

    this.room = room;
    this.synchronizedActions = synchronizedActions;
    this.synchronizedState = synchronizedState;
    
    this._socket = new WebSocket(serverUrl);
    this._isMaster = false;

    this._socket.onmessage = event => this.handleMessage(event.data);

    this.whenOpen = new Promise((resolve, reject) => {
      this._socket.onopen = () => resolve();
    });
  }

  get isPartiallySynchronizing() {
    return this.synchronizedActions.length > 0 || this.synchronizedState != null;
  }

  async synchronize(store) {
    if (store != null) {
      this.store = store;
    }

    await this.whenOpen;
    this._socket.send(JSON.stringify({ type: 'JOIN_ROOM', payload: this.room }));
  }

  loadInitialStateFromMaster() {
    const message = { type: 'REQUEST_STATE' };

    this._socket.send(JSON.stringify(message));
  }

  handleMessage(message) {
    const action = JSON.parse(message);

    switch (action.type) {
      case 'HEARTBEAT':
        this._isMaster = action.payload.isMaster;
        break;

      case 'JOIN_SUCCESSFUL':
        this._isMaster = action.payload.isMaster;
        this.dispatchEvent({ type: 'connect' });

        break;

      case 'REMOTE_ACTION':
        this.dispatchRemoteAction(action.payload);
        break;

      case 'REQUEST_STATE':
        this.sendLocalStateToServer(this.store);
        break;

      case 'SET_STATE':
        this.overwriteLocalState(action.payload);
        break;

      default:
        console.log('Warning: unknown message type received from server');
    }
  }

  dispatchRemoteAction(action) {
    if (this.isPartiallySynchronizing && !this.synchronizedActions.includes(action.type)) {
      return;
    }

    action[SERVER_MESSAGE_MARKER] = true;

    this.store.dispatch(action);
  }

  sendLocalStateToServer(store) {
    let payload = store.getState();

    if (this.isPartiallySynchronizing) {
      payload = extract(payload, this.synchronizedState);
    }

    const packet = { type: 'SET_STATE', payload };
    this._socket.send(JSON.stringify(packet));
  }

  sendLocalActionToServer(action) {
    if (this.isPartiallySynchronizing && !this.synchronizedActions.includes(action.type)) return;
    if (action.type === REDUX_INIT_ACTION || action.type === SET_STATE_ACTION) return;
    if (action[SERVER_MESSAGE_MARKER]) return;

    const packet = {
      type: 'SEND_ACTION',
      payload: action
    };

    this._socket.send(JSON.stringify(packet));
  }

  overwriteLocalState(newState) {
    if (this.isPartiallySynchronizing) {
      newState = extract(newState, this.synchronizedState, this.store.getState());
    }

    const action = {
      type: SET_STATE_ACTION,
      payload: newState
    };

    this.store.dispatch(action);
  }

  get middleware() {
    return store => next => action => {
      this.sendLocalActionToServer(action);

      const actionResult = next(action);

      if (this._isMaster) {
        this.sendLocalStateToServer(store);
      }

      return actionResult;
    };
  }

  get isMaster() {
    return this._isMaster;
  }

  get hasMaster() {
    return !this.isMaster;
  }
}

function createSynchronizedReducer(appReducer) {
  return (state, action) => {
    switch (action.type) {
      case SET_STATE_ACTION:
        return action.payload;
      default:
        return appReducer(state, action);
    }
  };
}

function isRemoteAction(action) {
  return action[SERVER_MESSAGE_MARKER] === true;
}

module.exports = { SyncClient, createSynchronizedReducer, isRemoteAction };
