const expect = require('expect');
const { runMiddleware, FAKE_ACTION } = require('./util/middlewareStub');

const { TestableSyncClient, createMasterClient, createNonMasterClient } = require('./util/TestableSyncClient');

describe('The synchronization middleware', function () {
  beforeEach(function () {
    TestableSyncClient.setUp();
  });

  afterEach(function () {
    TestableSyncClient.tearDown();
  });

  it('should be valid redux middleware', function () {
    expect(new TestableSyncClient().middleware).toBeReduxMiddleware();
  });

  it('should intercept all actions and sends them to the server', function () {
    const client = new TestableSyncClient();

    runMiddleware(client.middleware);
    expect(client).toHaveReportedLocalAction(FAKE_ACTION);
  });

  describe('when configured to sync partial state', function () {
    beforeEach(function () {
      this.client = new TestableSyncClient({
        synchronizedActions: ['ALLOWED_ACTION']
      });
    });

    it('should send whitelisted actions to the server normally', function () {
      const action = { type: 'ALLOWED_ACTION' };

      runMiddleware(this.client.middleware, { action });
      expect(this.client).toHaveReportedLocalAction(action);
    });

    it('should not send the server actions that are not in the whitelist', function () {
      const action = { type: 'FORBIDDEN_ACTION' };

      runMiddleware(this.client.middleware, { action });
      expect(this.client).toNotHaveReportedLocalAction(action);
    });
  });

  it('should not send the server actions that already originated from a remote source', function () {
    const client = new TestableSyncClient();
    const action = {
      type: 'FOOBAR_BAZ',
      ['@@FROM_SERVER']: true
    };
    
    runMiddleware(client.middleware, { action });
    expect(client).toNotHaveReportedLocalAction(action);
  });

  it ('should not send state initialization actions to the server', function () {
    const client = new TestableSyncClient();
    const specialActionTypes = ['@@INIT', '@@SET_STATE'];

    for (const type of specialActionTypes) {
      const action = { type };

      runMiddleware(client.middleware, { action });
      expect(client).toNotHaveReportedLocalAction(action);
    }
  });

  describe('when master', function () {
    it('should send a state backup to the server', function () {
      const client = createMasterClient();
      const fakeState = { message: 'Hello world', position:  {x: 0, y:0 } };

      runMiddleware(client.middleware, { state: fakeState });
      expect(client).toHaveSentState(fakeState);
    });
  });

  describe('when not master', function () {
    it('should not send a state backup to the server', function () {
      const client = createNonMasterClient();

      runMiddleware(client.middleware);
      expect(client).toNotHaveSentState();
    });
  });
});