const expect = require('expect');

require('./util/objectSimilarityExpectations');

const { TestableSyncClient, createClientInRoom, createClientSynchronizing } = require('./util/TestableSyncClient');

const FAKE_STATE = {
  message: 'HELLO_WORLD',
  interestingNumbers: [42, 1337, 0xDEADBEEF]
};

function createMockStore(storeData) {
  return {
    getState: () => storeData
  };
}

describe('Server events: ', function () {
  beforeEach(function () {
    TestableSyncClient.setUp();
  });

  afterEach(function () {
    TestableSyncClient.tearDown();
  });

  describe('HEARTBEAT', function () {
    it('should be able to notify the client that it is now the master', function () {
      const client = createClientInRoom('a-room');

      client.simulateHeartbeatReceived(true);
      expect(client.isMaster).toBe(true);
    });

    it('should be able to notify the client that it is no longer the master', function () {
      const client = createClientInRoom('a-room', true);

      client.simulateHeartbeatReceived(false);
      expect(client.isMaster).toBe(false);
    });
  });

  describe('JOIN_SUCCESSFUL', function () {
    it('should inform the client if it is the master', function () {
      const client = new TestableSyncClient();

      client.simulateJoinedRoomEvent(true);
      expect(client.isMaster).toBe(true);
    });

    it('should inform the client if it is not the master', function () {
      const client = new TestableSyncClient();

      client.simulateJoinedRoomEvent(false);
      expect(client.isMaster).toBe(false);
    });

    it('should cause the client to dispatch the connect event', function () {
      const client = new TestableSyncClient();

      const connectSpy = expect.createSpy();
      client.on('connect', connectSpy);
      
      client.simulateJoinedRoomEvent();
      expect(connectSpy).toHaveBeenCalled();
    });
  });

  describe('REMOTE_ACTION', function () {
    beforeEach(function () {
      this.store = {
        dispatch: expect.createSpy()
      };

      this.client = createClientSynchronizing(this.store);

      this.action = { type: 'SOMETHING_EXCITING_HAPPENED' };
      this.client.simulateRemoteActionReceived(this.action);
    });

    it('should dispatch the received action into the synchronized store', function () {
      expect(this.store.dispatch).toHaveBeenCalledWithSimilarObject(this.action);
    });

    it('should mark the received action to indicate that it originated from a remote source', function () {
      expect(this.store.dispatch).toHaveBeenCalledWithSimilarObject({
        ['@@FROM_SERVER']: true
      });
    });
  });

  describe('REQUEST_STATE', function () {
    it('should cause the client to send a backup of its state to the server', function () {
      const client = createClientSynchronizing(createMockStore(FAKE_STATE));

      client.simulateLocalStateRequested();
      expect(client).toHaveSentState(FAKE_STATE);
    });

    describe('when configured to sync partial state', function () {
      it('should only send whitelisted state to the server', function () {
        const entireState = {
          synchronizedReducer: 'some data',
          unsynchronizedReducer: 'other data'
        };

        const mockStore = createMockStore(entireState);
        const client = new TestableSyncClient({
          synchronizedState: ['synchronizedReducer']
        });

        client.synchronize(mockStore);
        client.simulateLocalStateRequested();

        expect(client).toHaveSentState({
          synchronizedReducer: 'some data'
        });
      });
    });
  });

  describe('SET_STATE', function () {
    it('should cause the client to override its state with the once received from the server', function () {
      const mockStore = {
        dispatch: expect.createSpy()
      };

      const client = createClientSynchronizing(mockStore);
      client.simulateRemoteStateReceived(FAKE_STATE);

      expect(mockStore.dispatch).toHaveBeenCalledWith({
        type: '@@SET_STATE',
        payload: FAKE_STATE
      });
    });
  });
});