const expect = require('expect');

const assert = require('../../node_modules/expect/lib/assert').default;
const { isEqual } = require('../../node_modules/expect/lib/TestUtils');

const { SyncClient } = require('../../src/syncClient');

class TestableSyncClient extends SyncClient {
  static setUp() {
    global.WebSocket = expect.createSpy();
    global.WebSocket.prototype.send = expect.createSpy();
  }

  static tearDown() {
    delete global.WebSocket;
  }

  simulateServerMessage(type, payload) {
    this._socket.onmessage({
      data: JSON.stringify({
        type, payload
      })
    });
  }

  simulateSocketConnected() {
    this._socket.onopen();
  }

  simulateJoinedRoomEvent(isMaster = false) {
    this.simulateServerMessage('JOIN_SUCCESSFUL', { isMaster });
  }

  simulateHeartbeatReceived(isMaster) {
    this.simulateServerMessage('HEARTBEAT', { isMaster });
  }

  simulateRemoteActionReceived(action) {
    this.simulateServerMessage('REMOTE_ACTION', action);
  }

  simulateLocalStateRequested() {
    this.simulateServerMessage('REQUEST_STATE');
  }

  simulateRemoteStateReceived(newState) {
    this.simulateServerMessage('SET_STATE', newState);
  }
}

function createMasterClient() {
  const client = new TestableSyncClient();
  client._isMaster = true;

  return client;
}

function createNonMasterClient() {
  const client = new TestableSyncClient();
  client._isMaster = false;

  return client;
}

function createClientInRoom(room, isMaster = false) {
  const client = new TestableSyncClient();
  if (room != null) {
    client.room = room;
  }

  client.simulateJoinedRoomEvent(isMaster);
  return client;
}

function createClientSynchronizing(store) {
  const client = new TestableSyncClient();

  client.synchronize(store);
  return client;
}

module.exports = {
  TestableSyncClient,
  createMasterClient,
  createNonMasterClient,
  createClientInRoom,
  createClientSynchronizing
};

expect.extend({
  toHaveSentPacket(type, payload) {
    const client = this.actual;

    assert(
      client instanceof TestableSyncClient,
      'The "client" argument in expect(client).ToHaveSentPacket() must be a TestableSyncClient'
    );

    assert(
      WebSocket.prototype.send.calls.some(call => isEqual(JSON.parse(call.arguments[0]).type, type)),
      'client did not send %s packet',
      type
    );

    if (payload !== undefined) {
      const expectedMessage = JSON.stringify({ type, payload });
      assert(
        WebSocket.prototype.send.calls.some(call => isEqual(call.arguments[0], expectedMessage)),
        'client did not send packet %s',
        type
      );
    }
  },

  toNotHaveSentPacket(type, payload) {
    const client = this.actual;

    assert(
      client instanceof TestableSyncClient,
      'The "client" argument in expect(client).ToHaveSentPacket() must be a TestableSyncClient'
    );

    assert(
      !WebSocket.prototype.send.calls.some(call => isEqual(JSON.parse(call.arguments[0]).type, type)),
      'client sent %s packet',
      type
    );

    if (payload !== undefined) {
      const expectedMessage = JSON.stringify({ type, payload });
      assert(
        !WebSocket.prototype.send.calls.some(call => isEqual(call.arguments[0], expectedMessage)),
        'client sent packet %s',
        type
      );
    }
  },

  toHaveReportedLocalAction(action) {
    expect(this.actual).toHaveSentPacket('SEND_ACTION', action);
  },

  toNotHaveReportedLocalAction(action) {
    expect(this.actual).toNotHaveSentPacket('SEND_ACTION', action);
  },

  toHaveSentState(localState) {
    expect(this.actual).toHaveSentPacket('SET_STATE', localState);
  },

  toNotHaveSentState() {
    expect(this.actual).toNotHaveSentPacket('SET_STATE');
  },

  toHaveJoinedRoom(roomName) {
    expect(this.actual).toHaveSentPacket('JOIN_ROOM', roomName);
  },

  toHaveRequestedRemoteState() {
    expect(this.actual).toHaveSentPacket('REQUEST_STATE');
  }
});
